# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package DB;

use strict;
use warnings FATAL => 'all';
use diagnostics;
use Pg;

require "sys/syscall.ph";

do "config.pl";

my $debug = defined($ENV{"QUERY_STRING"}) ? 0 : 1;

my $TIMEVAL_T = "LL";
my ($done, $start);
$done = $start = pack($TIMEVAL_T, ());
syscall(&SYS_gettimeofday, $start, 0);



$SIG{PIPE} = sub { print "Broken pipe!\n"; exit(0); };

my $option_ref = &Pg::conndefaults();
my $conn;

my $hostaddr;
my $dbname;
my $user;
my $password;


# 0: host
# 1: database name
# 2: username
# 3: password
sub set_db
{
	($hostaddr, $dbname, $user, $password) = @_;
}


sub connect
{
	my $cmd = "hostaddr=$hostaddr dbname='$dbname'" .
			" user=$user password=$password";
	$conn = &Pg::connectdb($cmd);
	($conn->status eq PGRES_CONNECTION_OK) || die $conn->errorMessage;
}


# 0: command
sub update
{
	my $query = $_[0];
	my $result;

	(defined($conn)) || &DB::connect();
	($debug == 1) && print "$query\n";
	$result = $conn->exec($query);
	if ($result->resultStatus ne PGRES_COMMAND_OK) {
		printf STDERR "wanha:db.pl:update update failed: $query\n";
		die $conn->errorMessage;
	}
}


# 0: command
sub query($)
{
	my $query = $_[0];
	my $result;

	(defined($conn)) || &DB::connect();
	($debug == 1) && print "$query\n";
	$result = $conn->exec($_[0]);
	if ($result->resultStatus ne PGRES_TUPLES_OK) {
		print STDERR "wanha:db.pl:query query failed: $query\n";
		die $conn->errorMessage;
	}

	return $result;
}


# 0: sequence name
# r: n
sub get_sequence
{
	my $sequence = $_[0];
	my ($cmd, $result, @row);

	(defined($conn)) || &DB::connect();
	$cmd = "SELECT nextval('$sequence') FROM $sequence";
	$result = &query($cmd);
	@row = $result->fetchrow;
	return $row[0];
}


sub transaction_start
{
	&update("BEGIN");
}


sub transaction_end
{
	&update("END");
}


# 0: string
# r: psql-safe string
sub safe_string
{
	$_ = $_[0];
	s/'//g;
	return $_;
}


sub process_time
{
	syscall(&SYS_gettimeofday, $done, 0);
	my @tstart = unpack($TIMEVAL_T, $start);
	my @tdone  = unpack($TIMEVAL_T, $done);

	$_ /= 1000000 for ($tdone[1], $tstart[1]);
	return ($tdone[0] + $tdone[1]) - ($tstart[0] + $tstart[1]);
}


return 1;
