# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package Chan;

use strict;
use warnings FATAL => 'all';
use diagnostics;

require "db.pl";


# 0: chan name
# r: chan id
sub add($$)
{
	my ($raw) = @_;
	my $id;
	my $query;

	my $name = &DB::safe_string($raw);

	&DB::transaction_start();
	$id = &DB::get_sequence("chans_id_seq");
	$query = "INSERT INTO chans (id, chan) VALUES ($id, '$name')";
	&DB::update($query);
	&DB::transaction_end();

	return $id;
}


# 0: chan name
# 1: auto-add (define to enable)
# r: chan id
sub get_id($$)
{
	my ($raw, $autoadd) = @_;
	my ($query, $result);

	my $name = &DB::safe_string($raw);

	$query = "SELECT id, chan FROM chans WHERE chan ILIKE '$name'";
	$result = &DB::query($query);

	if ($result->ntuples == 1) {
		my ($id, $t) = $result->fetchrow;
		return $id;
	} elsif ($result->ntuples == 0) {
		if (defined($autoadd) && $autoadd == 1) {
			return &add($name);
		}
		return;
	}

	# more than one entry? PostgreSQL/ILIKE seems to handle '_' as a
	# wildcard... Fine, we'll manually pick the correct one.
	$name =~ tr/A-Z/a-z/;
	while (my ($id, $str) = $result->fetchrow) {
		$str =~ tr/A-Z/a-z/;
		($str eq $name) && return $id;
	}
	return;
}


# 0: chan id / chan name
# r: chan id
sub get_id_smart($)
{
	my ($name) = @_;
	my $id;

	if ($name eq &Lang::get("txt-any")) {
		return -1;
	} elsif (! $name =~ /./) {
		return;
	} elsif ($name =~ /^[0-9]/) {
		return $name;
	} else {
		return &get_id($name);
	}
}


# 0: id
# r: chan
sub get_name($)
{
	my ($id) = @_;
	my $query;
	my $result;

	$id = &DB::safe_string($id);

	$query = "SELECT chan FROM chans WHERE id = $id";
	$result = &DB::query($query);
	($result->ntuples == 0) && return;

	my ($name) = $result->fetchrow;
	return $name;
}
