#! /usr/bin/perl
#
# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


use strict;
use warnings FATAL => 'all';
use diagnostics;
use Socket;
use Pg;

require "db.pl";
require "dbops.pl";
require "config.pl";

my $bot_pass = &Conf::bot_pass();
my $irc_debug = &Conf::irc_debug();
my $my_url = &Conf::base_url();
my $host = &Conf::irc_host();
my $port = &Conf::irc_port();
my $nick = &Conf::irc_nick();
my $cnick;
my $user = &Conf::irc_user();
my $real = &Conf::irc_real();
my $pass = &Conf::irc_pass();
my @chans = &Conf::irc_chans();

my $last_msg = time();
my $socket = -1;

$SIG{"INT"} = "graceful_exit";


if ($#ARGV >= 0) {
	&learn_logs();
	exit 0;
}



while (1) {
	($socket == -1) && &get_connected();
	($socket == -1) && next;

	while (<$socket>) {
		chop; chop;
		my $in = $_;
		($irc_debug & 1) && print "<< $in\n";

		my ($src, $cmd, $p0, $p1) = split(/ /, $in, 4);
		(defined($src) && ($irc_debug & 4)) && print "src: \"$src\"\n";
		(defined($cmd) && ($irc_debug & 4)) && print "cmd: \"$cmd\"\n";
		(defined($p0) && ($irc_debug & 4)) && print "p0 : \"$p0\"\n";
		(defined($p1) && ($irc_debug & 4)) && print "p1 : \"$p1\"\n";

		if ($cmd eq "PRIVMSG") {
			$p1 =~ s/^://;
			&parse_privmsg($p1, $src, $p0);
		} elsif ($cmd eq "433") {
			$cnick .= "_";
			&print_irc("NICK $cnick");
		} elsif ($cmd eq "003") {
			&join_channels();
		} elsif ($src eq "PING") {
			&print_irc("PONG $cmd");
		}
	}

	print "Disconnected!\n";
	sleep(10);
	$socket = -1;
}



sub get_connected
{
	my $proto = getprotobyname("tcp");
	((! defined($proto)) || (! $proto)) && return;
	socket(my $s, PF_INET, SOCK_STREAM, $proto);
	((! defined($s)) || (! $s)) && return;
	my $sin = sockaddr_in($port, inet_aton($host));
	((! defined($sin)) || (! $sin)) && return;
	my $r = connect($s, $sin);
	if ((! defined($r)) || ($r == -1)) {
		print "Can't connect!\n";
		sleep(10);
		$socket = -1;
	}
	$socket = $s;

	my $fh = select($socket);
	$| = 1;
	select($fh);

	(defined($pass)) && &print_irc("PASS $pass");
	&print_irc("NICK $nick");
	&print_irc("USER $user 0 0 :$real");

	$cnick = $nick;
}


sub join_channels
{
	foreach (@chans) {
		&print_irc("JOIN $_");
	}
}


# 0: message
# 1: src
# 2: target
sub parse_privmsg
{
	my ($msg, $src, $chan) = @_;
	my $url;

	($src) = $src =~ /^:(.*?)!/;

	if ($msg =~ /^$cnick[,:]? /i) {
		my ($stuff) = $msg =~ /^$cnick[,:]? (.*)/;
		if ($stuff =~ /^scores/i) {
			if ($chan eq $cnick) {
				&print_top($src);
			} else {
				&print_top($chan);
			}
			return;
		} elsif (($stuff =~ /^stfu$/i) || ($stuff =~ /^quiet$/i)) {
			&set_user_mute($src, $chan, 1);
		} elsif (($stuff =~ /^speak$/i) || ($stuff =~ /^verbose$/i)) {
			&set_user_mute($src, $chan, 0);
		} elsif ($stuff =~ /^learn $bot_pass: /i) {
			&teach_url($msg, $src, $chan);
			return;
		}
	}

	&parse_msg($msg, $src, $chan, undef, $chan);
}


# 0: message
# 1: src (who pasted)
# 2: chan (where pasted)
# 3: date (when pasted)
# 4: echo
sub parse_msg
{
	my ($msg, $src, $target, $date, $echo) = @_;
	my $url;
	my $bu_msg = $msg;

	while ($msg =~ /[a-z]+:\/\/[^\s]+/i) {
		my ($b) = $msg =~ /(.)?[a-z]+:\/\//i;
		my $del = " ";
		if ($b =~ /['"]/) {
			$del = $b;
		} elsif ($b =~ /\(/) {
			$del = ")";
		}
		$msg =~ s/\Q$del\E?([a-z]+:\/\/.*?)\Q$del\E(.*)/ $1 $2/i;
		($url, $msg) = $msg =~ /([a-z]+:\/\/[^\s]+)\s*(.*)/i;
		(defined($url)) || next;

		# remove full stop or comma, it hardly ever belongs there
		$url =~ s/[.,]$//;

		my $claim_w = ($bu_msg =~ /w[:!] \Q$url\E/) ? 1 : 0;
		&check_url($url, $target, $src, $date, $echo, $claim_w);
	}
}


# 0: msg
# 1: user
# 2: me
sub teach_url
{
	my ($msg, $user, $me) = @_;
	my ($chan, $date, $who, $line) = $msg =~
			/learn $bot_pass: (.*?) (.*?) <([^> ]+)> (.*)/;
	
	$date = &convert_date($date);
	
	&parse_msg($line, $who, $chan, $date, $user);

	print $socket "PRIVMSG $user :done\n";
}


# 0: url
# 1: where it was pasted
# 2: who pasted
# 3: time url was pasted
# 4: where to report if old
# 5: claim to be wanha
sub check_url
{
	my ($url, $chan_name, $nick_name, $time, $echo, $claim_w) = @_;

	($url =~ /^\Q$my_url\E/) && return;

	$chan_name =~ s/[^\#A-Z0-9\-.()\[\]~_^\%]/-/ig;
	$nick_name =~ s/[^\#A-Z0-9\-.()\[\]~_^\%]/-/ig;
	$url =~ s/(['"\x80-\xff])/sprintf("%%%02X", ord($1))/seg;

	my $chan_id = &DBOps::get_chan_id($chan_name, 1);
	my ($url_id, $url_count, $url_time, $nick_id) =
			&DBOps::get_url_info_by_url($url, $chan_id);
	
	if (defined($url_id) && $claim_w == 0) {
		# wanha with score
		my $wanha_id = &DBOps::get_nick_id($nick_name, 1);
		($wanha_id == $nick_id) && return;

		&wanha($url_id, $chan_id, $nick_id,
				$wanha_id, $nick_name, $echo);
	} elsif (! defined($url_id) && $claim_w == 1) {
		# cheating
		&cheat($nick_name, $chan_id, $echo);
	} elsif (defined($url_id) && $claim_w == 1) {
		# pasting known wanha
		# nick_id == original paster id
		# nick_name == wanha paster
		&known_wanha($url_id, $chan_id, $nick_id, $nick_name);
	}
	if (! defined($url_id)) {
		# pasted new url
		&add_url($url, $chan_id, $nick_name, $time);
	}
}


# 0: url id
# 1: chan id
# 2: original paste nick id
# 3: wanha paste nick id
# 4: wanha paste nick name
# 5: echo
sub wanha
{
	my ($url_id, $chan_id, $orig_id, $wanha_id, $wanha_name, $echo) = @_;
	my ($query, $result);
	my $t;

	&DB::connect();

	my ($orig_score, $wanha_score);

	# scores
	($orig_score, $t) = &DBOps::get_scoreboard($orig_id, $chan_id, 1);
	(defined($orig_score)) || die "wanha score";
	($t, $wanha_score) = &DBOps::get_scoreboard($wanha_id, $chan_id, 1);
	(defined($wanha_score)) || die "wanha wscore";

	# orig nick
	my $orig_name = &DBOps::get_nick_name($orig_id);

	# url info
	my ($url_count, $url_time) = &DBOps::get_url_info_by_id($url_id);

	# mute settings
	my $wanha_mute = &get_user_mute($wanha_id);
	my $orig_mute = &get_user_mute($orig_id);

	# announce
	$orig_score++;
	$wanha_score++;
	$url_count++;
	if (defined($echo) && $wanha_mute == 0 && $orig_mute == 0) {
		&print_irc("PRIVMSG $echo :$wanha_name, wanha! " .
			"$orig_name scores! " .
			"First pasted: $url_time", 1);
	} elsif (defined($echo) && $wanha_mute == 0 && $orig_mute == 1) {
		&print_irc("PRIVMSG $echo :$wanha_name, wanha! " .
			"First pasted: $url_time", 1);
	}

	# update score
	$query = "UPDATE scoreboard SET fame = $orig_score " .
			"WHERE nick = $orig_id AND chan = $chan_id";
	&DB::update($query);

	# update wanha score
	$query = "UPDATE scoreboard SET shame = $wanha_score " .
			"WHERE nick = $wanha_id AND chan = $chan_id";
	&DB::update($query);

	# update url
	$query = "UPDATE urls SET count = $url_count WHERE id = $url_id";
	&DB::update($query);
}


# 0: url id
# 1: chan id
# 2: original paster id
# 3: wanha paster nick
sub known_wanha
{
	my ($url_id, $chan_id, $orig_id, $wanha_nick) = @_;
	my ($t0, $t1, $t2);
	my $query;

	# url info
	my ($url_count) = &DBOps::get_url_info_by_id($url_id);
	(defined($url_count)) || die "known_wanha url_count";

	# wanha nick id
	my $wanha_id = &DBOps::get_nick_id($wanha_nick, 1);

	# scores
	my ($wanha_wanha, $orig_fame);
	($t0, $t1, $t2, $wanha_wanha) =
			&DBOps::get_scoreboard($wanha_id, $chan_id, 1);
	(defined($wanha_wanha)) || die "known_wanha wanha_id";
	($orig_fame) = &DBOps::get_scoreboard($orig_id, $chan_id, 1);
	(defined($orig_fame)) || die "known_wanha orig_id";

	# update url
	$url_count++;
	$query = "UPDATE urls SET count = $url_count WHERE id = $url_id";
	&DB::update($query);

	# update orig scoreboard
	$orig_fame++;
	$query = "UPDATE scoreboard SET fame = $orig_fame " .
			"WHERE nick = $orig_id AND chan = $chan_id";
	&DB::update($query);

	# update wanha scoreboard
	$wanha_wanha++;
	$query = "UPDATE scoreboard SET wanha = $wanha_wanha " .
			"WHERE nick = $wanha_id AND chan = $chan_id";
	&DB::update($query);
}


# 0: nick
# 1: chan id
# 2: chan name
sub cheat
{
	my ($nick_name, $chan_id, $chan_name) = @_;
	my ($query, $result);

	# cheater id
	my $nick_id = &DBOps::get_nick_id($nick_name, 1);
	(defined($nick_id)) || die "cheat nick";

	# scoreboard
	my ($fame, $shame, $cheat) =
			&DBOps::get_scoreboard($nick_id, $chan_id, 1);
	(defined($cheat)) || die "cheat score";
	$cheat++;

	# update score
	$query = "UPDATE scoreboard SET cheat = $cheat " .
			"WHERE nick = $nick_id AND chan = $chan_id";
	&DB::update($query);

	# announce
	(defined($chan_name)) || return;
	my $mute = &get_user_mute($nick_id);
	if ($mute == 0) {
		&print_irc("PRIVMSG $chan_name :$nick_name, cheater!", 1);
	}
}


# 0: url
# 1: chan id
# 2: nick name
# 3: timestamp
sub add_url
{
	my ($url, $chan_id, $nick_name, $time) = @_;

	my $nick_id = &DBOps::get_nick_id($nick_name, 1);
	(defined($nick_id)) || die "add_url nick";

	&DBOps::add_url($url, $chan_id, $nick_id, $time);
}


# 0: chan
sub print_top
{
	my ($chan) = @_;
	my ($query, $result);

	my $raw_chan = $chan;
	$chan =~ s/[^\#A-Z0-9\-.()\[\]~_^]/-/ig;

	my $top = 5;
	my $res;

	$query = "SELECT nicks.nick, scoreboard.fame " .
			"FROM nicks, scoreboard, chans " .
			"WHERE chans.chan = '$chan' " .
				"AND scoreboard.chan = chans.id " .
				"AND scoreboard.nick = nicks.id " .
			"ORDER BY scoreboard.fame DESC " .
			"LIMIT $top";
	$result = &DB::query($query);

	$res = "fame top $top:";
	while (my ($nick, $score) = $result->fetchrow) {
		$res .= " $nick: $score,";
	}
	$res =~ s/,$//;

	$query = "SELECT nicks.nick, scoreboard.shame " .
			"FROM nicks, scoreboard, chans " .
			"WHERE chans.chan = '$chan' " .
				"AND scoreboard.chan = chans.id " .
				"AND scoreboard.nick = nicks.id " .
			"ORDER BY scoreboard.shame DESC " .
			"LIMIT $top";
	$result = &DB::query($query);

	$res .= " - shame top $top:";
	while (my ($nick, $score) = $result->fetchrow) {
		$res .= " $nick: $score,";
	}
	$res =~ s/,$//;

	&print_irc("PRIVMSG $raw_chan :$res", 1);
}


# 0: nick
# 1: chan
# 2: mute (1 == mute)
sub set_user_mute
{
	my ($nick, $chan, $mute) = @_;
	my $id = &DBOps::get_nick_id($nick, 1);

	my $mute_s = $mute == 1 ? "TRUE" : "FALSE";

	my $query = "UPDATE nicks SET mute = $mute_s WHERE id = $id";
	&DB::update($query);

	print $socket "PRIVMSG $chan :ok\n";
}


# 0: nick id
# r: mute (1 == mute)
sub get_user_mute
{
	my ($id) = @_;
	my ($query, $result);

	$query = "SELECT mute FROM nicks WHERE id = $id";
	$result = &DB::query($query);
	($result->ntuples == 1) || die "user $id does not exist!";

	my ($mute) = $result->fetchrow;
	$mute = ($mute =~ /^t/i) ? 1 : 0;
	return $mute;
}


# 0: msg
# 1: safe
sub print_irc
{
	my ($msg, $safe) = @_;
	my $now = time();

	($now - $last_msg < 2 && defined($safe)) && return;
	($irc_debug & 2) && print ">> $msg\n";
	print $socket "$msg\n";
	$last_msg = $now;
}


sub graceful_exit
{
	($socket == -1) && exit(0);

	print $socket "QUIT :sigterm caught\n";
	sleep(1);
	exit(0);
}


# 0: date
sub convert_date
{
	my ($date) = @_;
	my %month = ( "Jan"=>"01", "Feb"=>"02", "Mar"=>"03", "Apr"=>"04",
			"May"=>"05", "Jun"=>"06", "Jul"=>"07", "Aug"=>"08",
			"Sep"=>"09", "Oct"=>"10", "Nov"=>"11", "Dec"=>"12" );
	
	my ($y, $m, $d, $t) = $date =~ /(\d+) (.*?) (\d+) (.*)/;
	return "$y-" . $month{"$m"} . "-$d $t";
}


sub learn_logs
{
	print "year? ";
	my $year = <STDIN>;
	print "channel? ";
	my $chan = <STDIN>;
	chop ($year, $chan);

	foreach (@ARGV) {
		open (FILE, "< $_") or die "Cannot read \"$_\"";
		while (<FILE>) {
			chop;
			my ($month, $day, $hour, $minute, $second,
					$who, $message) =
				/(.*?) (.*?) (\d+):(\d+):(\d+) <([^> ]+)> (.*)/;
			(defined($month)) || next;
			my $date = &convert_date("$year $month $day " .
					"$hour:$minute:$second");
			&parse_msg($message, $who, $chan, $date);
		}
		close (FILE);
	}
}
